;;;; -*- Mode:Lisp; Package:CL-USER; Syntax:COMMON-LISP; Base:10 -*-

(in-package :CL-USER)

;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
;;; Copyright (c) 1996 by Micheal Scott Hewett
;;;
;;; This code may be used by anyone for any project, but may not
;;; be sold in source or object form without permission.
;;; If in doubt, follow the GNU "copyleft" guidelines.
;;;
;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
;;; (Contact: hewett@cs.utexas.edu or hewett@cs.stanford.edu)
;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
;;;
;;;  User-level interaction (TELL and ASK) for the Algernon
;;;  Abstract Machine
;;;
;;;  (tell)
;;;  (ask)
;;;  (fire)
;;;  (verify)
;;;
;;;  (silently)
;;;  (with-no-continuations)
;;;  (deftaxonomy (taxonomy)...)
;;;
;;;  algy-note-loaded-kb
;;;
;;;  (acom-reset)
;;;  (algy-reset)
;;;  (a-assert)
;;;  (a-query)
;;;  (show-bindings)
;;;  (deftaxonomy taxonomy)
;;;
;;;  (find-rule <key>)
;;;
;;;  algy-trace
;;;  algy-trace-rule (rule-name)
;;;  algy-traced-rule-p (rule-name)
;;;  algy-untrace-rule (rule-name)
;;;  algy-untrace ()
;;;
;;;  algy-stats
;;;
;;;  ---------------------------------
;;;   7 Apr 1999 (mh) Added FIRE
;;;  30 Oct 1999 (mh) Added algy-stats and re-organized contents.
;;; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

;;; user-settable parameters

(defparameter *check-slot-domains*  T
  "Set to NIL to inhibit checking slot domains (after the application is
debugged).  The resulting application will run much faster.")



(defun TELL (path &KEY retrieve eval collect comment)
  "Asserts the clauses in the path.  Analyzes the results."

  (let (values-and-bindings)

    (when comment
      (terpri *trace-output*)
      (format *trace-output* comment))

    (when (>= aam::*trace-level* 2)
      (format t "~2%Tell: ~S" path))
    
    (let ((aam::*backward-chain* (not retrieve)))
      (setq values-and-bindings
	(aam:aam-process (aam:aam-compile path :ASSERT NIL) NIL))
      )


    (when (>= aam::*trace-level* 1)
      (aam::pp-output :TELL values-and-bindings))
    
    (aam::post-op   :TELL values-and-bindings :eval eval :collect collect)
    )
  )
	   

(defun ASK (path &KEY retrieve eval collect comment)
  "Queries the clauses in the path.  Analyzes the results."

  (let (values-and-bindings)

    (when comment
      (terpri *trace-output*)
      (format *trace-output* comment))

    (when (>= aam::*trace-level* 2)
      (format t "~2%Ask: ~S" path))
    
    (let ((aam::*backward-chain* (not retrieve)))
      (setq values-and-bindings
	(aam:aam-process (aam:aam-compile path :QUERY NIL) NIL))
      )

;;    (setq return-value (not (null values-and-bindings)))

    (when (>= aam::*trace-level* 1)
      (aam::pp-output :ASK (copy-tree values-and-bindings))
      )
    
    (aam::post-op   :ASK values-and-bindings :eval eval :collect collect)
    )
  )


(defun FIRE (clause &KEY retrieve eval collect comment)
  "Fires rules as if CLAUSE had just been asserted,
except that the (topmost) rules ALWAYS fire even if
CLAUSE is already in the KB.
This function should be used with caution.  It is useful
after CREATE-CLASS has been used to create a frame because
CREATE-CLASS adds slots to the frame without firing rules.
It does not check to see whether CLAUSE is ground or whether
it is actually in the KB."

  (let (values-and-bindings)

    (when comment
      (terpri *trace-output*)
      (format *trace-output* comment))

    (when (>= aam::*trace-level* 2)
      (format t "~2%Fire: ~S" clause))
    
    (let ((aam::*backward-chain* (not retrieve)))
      (setq values-and-bindings
	(aam:aam-process (aam:aam-compile `((:FIRE ,clause)) :ASSERT NIL) NIL))
      )

    (when (>= aam::*trace-level* 1)
      (aam::pp-output :TELL values-and-bindings))
    
    (aam::post-op   :TELL values-and-bindings :eval eval :collect collect)
    )
  )




(defun VERIFY (path &KEY retrieve eval collect comment)
  "Queries the clauses in the path.  Analyzes the results."

  (let (values-and-bindings
	(return-value NIL))

    (when comment
      (terpri *trace-output*)
      (format *trace-output* comment))

    (when (>= aam::*trace-level* 2)
      (format t "~2%VERIFY: ~S" path))
    
    (let ((aam::*backward-chain* (not retrieve)))
      (setq values-and-bindings
	(aam:aam-process (aam:aam-compile path :VERIFY NIL) NIL))
      )

    (setq return-value (not (null values-and-bindings)))

    (when (>= aam::*trace-level* 1)
      (aam::pp-output :VERIFY values-and-bindings)
      (aam::post-op   :VERIFY values-and-bindings :eval eval :collect collect))

    return-value
    )
  )



  
(defun show-bindings (binding-set)
  "Prints a binding-set in a nice way."

  (dolist (binding-list binding-set)
    (format t "~2%   Bindings:")
    (dolist (binding binding-list)
      (format t "~%~12@A = ~A" (car binding) (cdr binding))
      ))
  )



;;; ---------  Handling KBs - loading, reloading, resetting  ---------

(defparameter *ALGY-LOADED-KBS*  NIL
  "Records all KBs loaded in.")
    
(defun algy-note-loaded-kb (kb)
  "Adds a KB to the list of KBs that should be automatically
loaded when (acom-reset) is called.  Can be a string or
a keyword denoting an internal kb."
  
  ;;; Must insert them in the same order as loaded.
  (unless (member kb *ALGY-LOADED-KBS*)
    (if (null *ALGY-LOADED-KBS*)
	(setq *ALGY-LOADED-KBS* (list kb))
      ;;ELSE
        (nconc *ALGY-LOADED-KBS* (list kb)))))


;;; ---------  USEFUL Algy user-level functions  ---------

;;; This is a commonly-called function in Algernon systems.
;;; This should keep a list of loaded KBs and load all of them.

(defun acom-reset ()

  (setq aam::*TOP-LEVEL-EXIT-TAG* NIL)    ;; Reset Catch/Throw mechanism.

  (sfs::kb-reset)

  (dolist (kb *ALGY-LOADED-KBS*)
    (if (stringp kb)
	(load kb)
      ;;else
      (case kb
	(:CORE        (load-algy-file "algy-core-kb"))
	(:COMMON      (load-algy-file "algy-common-kb"))
	(:SET-THEORY  (load-algy-file "algy-set-theory-kb"))
	)
      )
    )
  )


(defun algy-reset ()
  "Same as ACOM-RESET."
  (acom-reset)
  )


(defun A-ASSERT (sentence predicates &key collect)
  "Same as TELL surrounded by with-normal-output."
  
  (with-normal-output
    (tell predicates :comment sentence :collect collect)))

(defun A-QUERY (sentence predicates &key collect)
  "Same as ASK surrounded by with-normal-output."

  (with-normal-output
    (ask predicates :comment sentence :collect collect)))



(defmacro WITH-NO-CONTINUATIONS (&REST body)

  `(let ((aam::*make-continuation* NIL))
     ,@body)
  )


(defmacro deftaxonomy (taxonomy)
  "Same as the :TAXONOMY special form in Algernon."

  `(:taxonomy ',taxonomy))


;;;--------- User-level access functions

;; Modified 06 Feb 1998 (mh) to return a list of rules
;; found rather than printing them out.  This lets
;; the function be used programmatically.
(defun find-rule (key)
  "Given the 'key' clause, locates the rule or rules."

  (let ((rules-found NIL))
    (dolist (rule (mapcar #'car (sfs::kb-get-values 'rules 'member 'value)))
      (when (equal (car (sfs::kb-get-values rule 'key 'value)) key)
	(push rule rules-found)
	;; (format *standard-output* "~%~S" rule))
	)
      )
    rules-found
    )
  )


(defun algy-trace (&OPTIONAL (level 1))
  "0: no tracing
1: normal operation success/failure
2: extended trace - shows rules being activated and executed
3: additionally shows internal registers."

  (aam::trace-aam level)
  )


(defmacro SILENTLY (&REST body)
  "Sets trace level to 0 while executing the body."

  `(with-aam-silent
     ,@body
     )
  )


(defun algy-trace-rule (rule-name)
  "Sets trace level to 2 when the given rule is being executed."

  (aam::trace-rule rule-name)
  )


(defun algy-traced-rule-p (rule-name)
  "Returns T if the rule is being traced."

  (aam::rule-traced-p rule-name)
  )

(defun algy-untrace-rule (rule-name)
  "No tracing for the given rule."

  (aam::untrace-rule rule-name)
  )


(defun algy-untrace ()
  "Removes all rule tracing and sets the trace level to 1."

  (aam::untrace-rules)
  (algy-trace 1)
  )

(defun algy-stats (&OPTIONAL (stream *standard-output*))
  "Counts number of frames, slots and relations in the current KB.
If STREAM is non-NIL, prints to the stream.
Always returns the count of frames and relations in a keyword list."
  
  (let ((frames     (length (sfs:kb-get-all-frames)))
	(relations  (reduce #'+
			    (mapcar
			     #'(lambda (frame)
				 (count-relations-of-frame frame)
				 )
			     (sfs::kb-get-all-frames))
			    ))
	)

    (when stream
      (format stream "~%Algernon stats  @  ~A" (algy-get-time-string t))
      (format stream "~%     Frames: ~7D" frames)
      (format stream "~%  Relations: ~7D" relations)
      (terpri stream)
      )
  
    `(:frames ,frames :relations ,relations)
    )
  )
     
     
(defun count-relations-of-frame (frame)

  (let ((count  0))
    
    (dolist (slot-info (get frame 'SFS::slot-info))

      (dolist (facet-info (cdr slot-info))
	  ;; Special handling for rule slots to get them to print nicer.

	(incf count (length (cdr facet-info)))
	)
      )
    count
    )
  )

(defun algy-get-time-string (&OPTIONAL (short-version? nil))
  "Interprets (get-decoded-time) and returns a string with the date and time in some nice format."

  ;; Cobbled from old BB1 code.
  
  (multiple-value-bind (second minute hour date month year day-of-week ignore1 ignore2)           ;; Port.MHewett
      (get-decoded-time)
    ignore1 ignore2                                                                               ;; Port.Misc  
    (if short-version?
	(format NIL "~D ~A ~D  ~D:~2,'0D"  ;1 January 1987 16:14"
		date 
		(cdr (assoc month '((1 . "January")    (2 . "February") (3 . "March")     (4 . "April")
				    (5 . "May")        (6 . "June")     (7 . "July")      (8 . "August")
				    (9 . "September") (10 . "October") (11 . "November") (12 . "December"))))
		year
		hour
		minute)
	
	;;Else give long verbose date and time
	
	(format NIL "~A, ~D ~A ~D  ~D:~2,'0D:~2,'0D"	   ;Sunday, 1 January 1986  16:14:12
		
		(cdr (assoc day-of-week '((0 . "Monday") (1 . "Tuesday")  (2 . "Wednesday") (3 . "Thursday")
					  (4 . "Friday") (5 . "Saturday") (6 . "Sunday"))))
		date
		(cdr (assoc month '((1 . "January")    (2 . "February") (3 . "March")     (4 . "April")
				    (5 . "May")        (6 . "June")     (7 . "July")      (8 . "August")
				    (9 . "September") (10 . "October") (11 . "November") (12 . "December"))))
		year
		hour
		minute
		second))))
